"Test replace()."

import numpy as np
from numpy.testing import assert_equal, assert_array_equal, assert_raises
nan = np.nan
import bottlechest as bn


def arrays(dtypes=bn.dtypes, nans=True):
    "Iterator that yield arrays to use for unit testing."
    ss = {}
    ss[0] = {'size':  0, 'shapes': [(0,), (0,0), (2,0), (2,0,1)]}
    ss[1] = {'size':  4, 'shapes': [(4,)]}
    ss[2] = {'size':  6, 'shapes': [(1,6), (2,3)]}
    ss[3] = {'size':  6, 'shapes': [(1,2,3)]}
    ss[4] = {'size': 24, 'shapes': [(1,2,3,4)]}  # Unaccelerated
    for ndim in ss:
        size = ss[ndim]['size']
        shapes = ss[ndim]['shapes']
        for dtype in dtypes:
            a = np.arange(size, dtype=dtype)
            for shape in shapes:
                a = a.reshape(shape)
                yield a
            if issubclass(a.dtype.type, np.inexact): 
                if nans:
                    for i in range(a.size):
                        a.flat[i] = np.nan
                        yield a
                for i in range(a.size):
                    a.flat[i] = np.inf
                    yield a
    if nans:
        # nanmedian regression tests
        a = np.array([1, nan, nan, 2])
        yield a
        a = np.vstack((a, a))
        yield a
        yield a.reshape(1,2,4)

def unit_maker(func, func0, nans=True):
    "Test that bn.xxx gives the same output as np.xxx."
    msg = '\nfunc %s | input %s (%s) | shape %s | old %f | new %f\n'
    msg += '\nInput array:\n%s\n'
    olds = [0, np.nan, np.inf]
    news = [1, 0, np.nan]
    for i, arr in enumerate(arrays(nans=nans)):
        for old in olds:
            for new in news:
                if not issubclass(arr.dtype.type, np.inexact):
                    if not np.isfinite(old):
                        # Cannot safely cast to int
                        continue
                    if not np.isfinite(new):
                        # Cannot safely cast to int
                        continue
                actual = arr.copy()
                func(actual, old, new)
                desired = arr.copy()
                func0(desired, old, new)
                tup = (func.__name__, 'a'+str(i), str(arr.dtype),
                       str(arr.shape), old, new, arr)
                err_msg = msg % tup
                assert_array_equal(actual, desired, err_msg=err_msg)
                err_msg += '\n dtype mismatch %s %s'
                if hasattr(actual, 'dtype') or hasattr(desired, 'dtype'):
                    da = actual.dtype
                    dd = desired.dtype
                    assert_equal(da, dd, err_msg % (da, dd))

def test_replace():
    "Test replace."
    yield unit_maker, bn.replace, bn.slow.replace

# ---------------------------------------------------------------------------
# Check that exceptions are raised

def test_replace_unsafe_cast():
    "Test replace for unsafe casts."
    dtypes = [x for x in bn.dtypes if 'int' in x]
    shapes = [(0,), (2,0), (1,2,0)]
    for shape in shapes:
        for dtype in dtypes:
            a = np.zeros(shape, dtype=dtype)
            assert_raises(ValueError, bn.replace, a, 0.1, 0)
            assert_raises(ValueError, bn.replace, a, 0, 0.1)
            assert_raises(ValueError, bn.slow.replace, a, 0.1, 0)
            assert_raises(ValueError, bn.slow.replace, a, 0, 0.1)

def test_non_array():
    "Test that non-array input raises"
    a = [1, 2, 3]
    assert_raises(TypeError, bn.replace, a, 0, 1)
    a = (1, 2, 3)
    assert_raises(TypeError, bn.replace, a, 0, 1)

# ---------------------------------------------------------------------------
# Make sure bn.replace and bn.slow.replace can handle int arrays where
# user wants to replace nans

def test_replace_nan_int():
    "Test replace, int array, old=nan, new=0"
    a = np.arange(2*3*4).reshape(2,3,4)
    actual = a.copy()
    bn.replace(actual, np.nan, 0)
    desired = a.copy()
    msg = 'replace failed on int input looking for nans'
    assert_array_equal(actual, desired, err_msg=msg)
    actual = a.copy()
    bn.slow.replace(actual, np.nan, 0)
    msg = 'slow.replace failed on int input looking for nans'
    assert_array_equal(actual, desired, err_msg=msg)



"countnans template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["countnans"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = "float64"
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
                              np.ndarray[np.float64_t, ndim=1] weights=None):
    "Count of nans in NDIMd array with dtype=DTYPE along axis=AXIS."
    cdef np.DTYPE_t ai
"""

loop = {}
loop[2] = """
    cdef float fcnt
    cdef int count
    if weights is not None:
        for iINDEX0 in range(nINDEX0):
            fcnt = 0
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    fcnt += weights[iINDEX1]
            y[INDEXPOP] = fcnt
    else:
        for iINDEX0 in range(nINDEX0):
            count = 0
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    count += 1
            y[INDEXPOP] = count
    return y"""

floats['loop'] = loop

# Float dtypes (axis=None) --------------------------------------------------

floats_None = deepcopy(floats)
floats_None['axisNone'] = True

loop = {}
loop[1] = """
    if weights is not None:
        if weights.ndim != 1 or weights.shape[0] != nINDEX0:
             raise ValueError("shape of weights does not match the data")
        fcnt = 0
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai != ai:
                fcnt += weights[iINDEX0]
        return fcnt
    else:
        count = 0
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai != ai:
                count += 1
        return count"""

loop[2] = """
    if weights is not None:
        if weights.shape[0] != (nINDEX0, nINDEX1):
             raise ValueError("shape of weights does not match the data")
        fcnt = 0
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    fcnt += weights[INDEXALL]
        return fcnt
    else:
        count = 0
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    count += 1
        return count"""

floats_None['loop'] = loop

# Int dtypes (not axis=None) ------------------------------------------------


# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "countnans"
slow['signature'] = "arr, weights"
slow['func'] = "bn.slow.countnans(arr, weights, axis=AXIS)"

# Template ------------------------------------------------------------------

countnans = {}
countnans['name'] = 'countnans'
countnans['is_reducing_function'] = True
countnans['cdef_output'] = True
countnans['slow'] = slow
countnans['templates'] = {}
countnans['templates']['float'] = floats
countnans['templates']['float_None'] = floats_None
countnans['pyx_file'] = 'func/%sbit/countnans.pyx'

countnans['main'] = '''"countnans auto-generated from template"

def countnans(arr, weights=None, axis=None):
    """
    Count the undefined elements along given axis.

    Parameters
    ----------
    arr : array_like
        Array in which nans are to be counted. If `arr` is not an array, a
        conversion is attempted.
    weights : {array-like, None}
        Array containing the weights along the specified axis.
        Note: weights, if present, must be of type `float64`
    axis : {int, None}, optional
        Axis along which the nans are counted. The default (axis=None) is to
        count them across the entire array.

    Returns
    -------
    y : ndarray
        An array with the same shape as `arr`, with the specified axis removed.
        The output array is an array of float64 if data is weighted, or of int,
        if it is not.

    Examples
    --------
    >>> bn.countnans(1)
    0.0
    >>> bn.countnans([1])
    0.0
    >>> bn.countnans([1, np.nan])
    1.0
    >>> a = np.array([[1, 4], [1, np.nan]])
    >>> bn.countnans(a)
    1.0
    >>> bn.countnans(a, axis=0)
    array([ 0, 1])

    """
    func, arr, weights = countnans_selector(arr, weights, axis)
    return func(arr, weights)

def countnans_selector(arr, weights, axis):
    """
    Return countnans function and array that matches `arr` and `axis`.

    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in
    bn.countnans() is in checking that `axis` is within range, converting `arr`
    into an array (if it is not already an array), and selecting the function
    to use to count the nans.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    weights : {array-like, None}
        Array containing the weights along the specified axis.
        Note: weights, if present, must be of type `float64`
    axis : {int, None}
        Axis along which the nans are to be counted.

    Returns
    -------
    func : function
        The countnans function that matches the number of dimensions and dtype
        of the input array and the axis along which you wish to count the nans.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.
    weights : {array-like, None}
        If the input array `weights` is not a ndarray, then `weights` will
        contain the result of converting `weights` into a ndarray.

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:
        a = np.array(arr, copy=False)
    if weights is not None and type(weights) is not np.ndarray:
        weights = np.array(weights, copy=False)
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if (axis is not None) and (axis < 0):
        axis += ndim
    cdef tuple key = (ndim, dtype, axis)
    try:
        if weights is not None and weights.ndim > 1:
            raise KeyError # fast function supports only 1-dimensional weights
        func = countnans_dict[key]
    except KeyError:
        if axis is not None:
            if (axis < 0) or (axis >= ndim):
                raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = countnans_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a, weights
'''



from __future__ import absolute_import

# Supported dtypes
dtypes = ['int', 'int8', 'int32', 'int64', 'float', 'float32', 'float64']

from . import slow

# If you bork the build (e.g. by messing around with the templates),
# you still want to be able to import Bottleneck so that you can
# rebuild using the templates. So try to import the compiled Bottleneck
# functions to the top level, but move on if not successful.
try:
    from . import func
    from .allnan import allnan
    from .nansum import nansum
    from .nanmax import nanmax
    from .nanmin import nanmin
    from .nanmean import nanmean
    from .nanstd import nanstd
    from .nanvar import nanvar
    from .median import (median, nanmedian)
    from .nanargmin import nanargmin
    from .nanargmax import nanargmax
    from .rankdata import (rankdata, nanrankdata)
    from .ss import ss
    from .nn import nn
    from .replace import replace
    from .anynan import anynan
    from .allnan import allnan
    from .nanequal import nanequal
    from .bincount import bincount
    from .valuecount import valuecount
    from .countnans import countnans
    from .contingency import contingency
    from .stats import stats
    from .partsort import partsort
    from .argpartsort import argpartsort
except ImportError:
    pass


from bottlechest.version import __version__
from bottlechest.benchmark.bench import bench

try:
    from numpy.testing import Tester
    test = Tester().test
    del Tester
except (ImportError, ValueError):
    print("No Bottleneck unit testing available.")



"Test functions."

# For support of python 2.5
from __future__ import with_statement

import unittest
import numpy as np
from numpy.testing import (assert_equal, assert_array_equal, assert_raises,
                           assert_array_almost_equal, assert_almost_equal)
nan = np.nan
import bottlechest as bn


def arrays(dtypes=bn.dtypes, nans=True):
    "Iterator that yields arrays to use for unit testing."
    ss = {}
    ss[0] = {'size':  0, 'shapes': [(0,), (0,0), (2,0), (2,0,1)]}
    ss[1] = {'size':  4, 'shapes': [(4,)]}
    ss[2] = {'size':  6, 'shapes': [(1,6), (2,3)]}
    ss[3] = {'size':  6, 'shapes': [(1,2,3)]}
    ss[4] = {'size': 24, 'shapes': [(1,2,3,4)]}  # Unaccelerated
    for ndim in ss:
        size = ss[ndim]['size']
        shapes = ss[ndim]['shapes']
        for dtype in dtypes:
            a = np.arange(size, dtype=dtype)
            for shape in shapes:
                a = a.reshape(shape)
                yield a
                yield -a
            if issubclass(a.dtype.type, np.inexact): 
                if nans:
                    for i in range(a.size):
                        a.flat[i] = np.nan
                        yield a
                        yield -a
                for i in range(a.size):
                    a.flat[i] = np.inf
                    yield a
                    yield -a
    if nans:
        # nanmedian regression tests
        a = np.array([1, nan, nan, 2])
        yield a
        a = np.vstack((a, a))
        yield a
        yield a.reshape(1,2,4)

def unit_maker(func, func0, decimal=np.inf, nans=True):
    "Test that bn.xxx gives the same output as bn.slow.xxx."
    msg = '\nfunc %s | input %s (%s) | shape %s | axis %s\n'
    msg += '\nInput array:\n%s\n'
    for i, arr in enumerate(arrays(nans=nans)):
        for axis in list(range(-arr.ndim, arr.ndim)) + [None]:
            with np.errstate(invalid='ignore'):
                actual = 'Crashed'
                desired = 'Crashed'
                actualraised = False
                try:
                    actual = func(arr.copy(), axis=axis)
                except:
                    actualraised = True
                desiredraised = False
                try:
                    desired = func0(arr.copy(), axis=axis)
                except:
                    desiredraised = True
            if actualraised and desiredraised:
                pass
            else:
                tup = (func.__name__, 'a'+str(i), str(arr.dtype),
                       str(arr.shape), str(axis), arr)
                err_msg = msg % tup
                if (decimal < np.inf) and (np.isfinite(arr).sum() > 0):
                    assert_array_almost_equal(actual, desired, decimal, err_msg)
                else:
                    assert_array_equal(actual, desired, err_msg)
                err_msg += '\n dtype mismatch %s %s'
                if hasattr(actual, 'dtype') or hasattr(desired, 'dtype'):
                    da = actual.dtype
                    dd = desired.dtype
                    assert_equal(da, dd, err_msg % (da, dd))

def test_nansum():
    "Test nansum."
    yield unit_maker, bn.nansum, bn.slow.nansum

def test_nanmax():
    "Test nanmax."
    yield unit_maker, bn.nanmax, bn.slow.nanmax

def test_nanargmin():
    "Test nanargmin."
    yield unit_maker, bn.nanargmin, bn.slow.nanargmin

def test_nanargmax():
    "Test nanargmax."
    yield unit_maker, bn.nanargmax, bn.slow.nanargmax

def test_nanmin():
    "Test nanmin."
    yield unit_maker, bn.nanmin, bn.slow.nanmin

def test_nanmean():
    "Test nanmean."
    yield unit_maker, bn.nanmean, bn.slow.nanmean, 5

def test_nanstd():
    "Test nanstd."
    yield unit_maker, bn.nanstd, bn.slow.nanstd, 5

def test_nanvar():
    "Test nanvar."
    yield unit_maker, bn.nanvar, bn.slow.nanvar, 5

def test_median():
    "Test median."
    yield unit_maker, bn.median, bn.slow.median, np.inf, False

def test_nanmedian():
    "Test nanmedian."
    yield unit_maker, bn.nanmedian, bn.slow.nanmedian
   
def test_rankdata():
    "Test rankdata."
    yield unit_maker, bn.rankdata, bn.slow.rankdata

def test_nanrankdata():
    "Test nanrankdata."
    yield unit_maker, bn.nanrankdata, bn.slow.nanrankdata

def test_ss():
    "Test ss."
    yield unit_maker, bn.ss, bn.slow.ss

def test_anynan():
    "Test anynan."
    yield unit_maker, bn.anynan, bn.slow.anynan

def test_allnan():
    "Test allnan."
    yield unit_maker, bn.allnan, bn.slow.allnan

# ---------------------------------------------------------------------------
# Check that exceptions are raised

def test_nanmax_size_zero(dtypes=bn.dtypes):
    "Test nanmax for size zero input arrays."
    shapes = [(0,), (2,0), (1,2,0)]
    for shape in shapes:
        for dtype in dtypes:
            a = np.zeros(shape, dtype=dtype)
            assert_raises(ValueError, bn.nanmax, a)
            assert_raises(ValueError, bn.slow.nanmax, a)
            
def test_nanmin_size_zero(dtypes=bn.dtypes):
    "Test nanmin for size zero input arrays."
    shapes = [(0,), (2,0), (1,2,0)]
    for shape in shapes:
        for dtype in dtypes:
            a = np.zeros(shape, dtype=dtype)
            assert_raises(ValueError, bn.nanmin, a)
            assert_raises(ValueError, bn.slow.nanmin, a)

# ---------------------------------------------------------------------------
# Check nn

def arrays2(dtypes=bn.dtypes):
    "Iterator that yield arrays to use for unit testing."
    for dtype in bn.dtypes:
        arr = np.random.randint(0, 10, (2, 3)).astype(dtype)
        arr0 = np.random.randint(0, 10, 3).astype(dtype)
        axis = 0
        yield arr.copy(), arr0[:2].copy(), axis
        axis = 1
        yield arr.copy(), arr0.copy(), axis
        # Make sure ties are handled in the same way
        arr.fill(0)
        arr0.fill(0)
        axis = 0
        yield arr.copy(), arr0[:2].copy(), axis
        axis = 1
        yield arr.copy(), arr0.copy(), axis
        if issubclass(arr.dtype.type, np.inexact): 
            # Make sure NaNs are handled in the same way
            arr.fill(np.nan)
            arr0.fill(np.nan)
            axis = 0
            yield arr.copy(), arr0[:2].copy(), axis
            axis = 1
            yield arr.copy(), arr0.copy(), axis

def test_nn():
    "Test that bn.nn gives the same output as bn.slow.nn."
    msg = '\nfunc %s | input %s (%s) | axis %s\n'
    msg += '\nInput array `arr`:\n%s\n'
    msg += '\nInput array `arr0`:\n%s\n'
    count = -1
    for arr, arr0, axis in arrays2():
        actual = bn.nn(arr, arr0, axis)
        desired = bn.slow.nn(arr, arr0, axis)
        count += 1
        tup = ('bn.nn', 'arr'+str(count), str(arr.dtype),
               str(axis), arr, arr0)
        err_msg = msg % tup
        assert_almost_equal(actual, desired, decimal=5, err_msg=err_msg)



"median template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["median", "nanmedian"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# loops ---------------------------------------------------------------------

loop = {}
loop[1] = """\
    if nINDEX0 == 0:
        return np.FLOAT(NAN)
    k = nAXIS >> 1
    l = 0
    r = nAXIS - 1 
    with nogil:       
        while l < r:
            x = b[k]
            i = l
            j = r
            while 1:
                while b[i] < x: i += 1
                while x < b[j]: j -= 1
                if i <= j:
                    tmp = b[i]
                    b[i] = b[j]
                    b[j] = tmp
                    i += 1
                    j -= 1
                if i > j: break
            if j < k: l = i
            if k < i: r = j
    if nAXIS % 2 == 0:        
        amax = MINDTYPE
        for i in range(k):
            ai = b[i]
            if ai >= amax:
                amax = ai
        return np.FLOAT(0.5 * (b[k] + amax))
    else:
        return np.FLOAT(b[k])
"""        
loop[2] = """\
    if nINDEX1 == 0:
        PyArray_FillWithScalar(y, NAN)
        return y
    for iINDEX0 in range(nINDEX0): 
        k = nAXIS >> 1
        l = 0
        r = nAXIS - 1
        while l < r:
            x = b[INDEXREPLACE|k|]
            i = l
            j = r
            while 1:
                while b[INDEXREPLACE|i|] < x: i += 1
                while x < b[INDEXREPLACE|j|]: j -= 1
                if i <= j:
                    tmp = b[INDEXREPLACE|i|]
                    b[INDEXREPLACE|i|] = b[INDEXREPLACE|j|]
                    b[INDEXREPLACE|j|] = tmp
                    i += 1
                    j -= 1
                if i > j: break
            if j < k: l = i
            if k < i: r = j
        if nAXIS % 2 == 0:        
            amax = MINDTYPE
            for i in range(k):
                ai = b[INDEXREPLACE|i|]
                if ai >= amax:
                    amax = ai
            y[INDEXPOP] = 0.5 * (b[INDEXREPLACE|k|] + amax)
        else:
            y[INDEXPOP] = CASTb[INDEXREPLACE|k|]         
    return y
"""

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = False
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Median of NDIMd array with dtype=DTYPE along axis=AXIS."
    cdef np.npy_intp i, j = 0, l, r, k 
    cdef np.DTYPE_t x, tmp, amax, ai
    cdef np.ndarray[np.DTYPE_t, ndim=NDIM] b = PyArray_Copy(a)
"""

floats['loop'] = {}
floats['loop'][1] = loop[1].replace('FLOAT', 'DTYPE')
floats['loop'][2] = loop[2].replace('CAST', '')

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 
ints['force_output_dtype'] = 'float64'
ints['loop'] = {}
ints['loop'][1] = loop[1].replace('FLOAT', 'float64')
ints['loop'][2] = loop[2].replace('CAST', '<np.float64_t> ')

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "median"
slow['signature'] = "arr"
slow['func'] = "bn.slow.median(arr, axis=AXIS)"

# Template ------------------------------------------------------------------

median = {}
median['name'] = 'median'
median['is_reducing_function'] = True
median['cdef_output'] = True
median['slow'] = slow
median['templates'] = {}
median['templates']['float'] = floats
median['templates']['int'] = ints
median['pyx_file'] = 'func/%sbit/median.pyx'

median['main'] = '''"median auto-generated from template"
# Select smallest k elements code used for inner loop of median method:
# http://projects.scipy.org/numpy/attachment/ticket/1213/quickselect.pyx
# (C) 2009 Sturla Molden
# SciPy license 
#
# From the original C function (code in public domain) in:
#   Fast median search: an ANSI C implementation
#   Nicolas Devillard - ndevilla AT free DOT fr
#   July 1998
# which, in turn, took the algorithm from
#   Wirth, Niklaus
#   Algorithms + data structures = programs, p. 366
#   Englewood Cliffs: Prentice-Hall, 1976
#
# Adapted and expanded for Bottleneck:
# (C) 2010 Keith Goodman

def median(arr, axis=None):
    """
    Median of array elements along given axis.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}, optional
        Axis along which the median is computed. The default (axis=None) is to
        compute the median of the flattened array.

    Returns
    -------
    y : ndarray
        An array with the same shape as `arr`, except that the specified axis
        has been removed. If `arr` is a 0d array, or if axis is None, a scalar
        is returned. `float64` return values are used for integer inputs. 
    
    See also
    --------
    bottlechest.nanmedian: Median along specified axis ignoring NaNs. 

    Notes
    -----
    This function returns the same output as NumPy's median except when the
    input contains NaN.

    Examples
    --------
    >>> a = np.array([[10, 7, 4], [3, 2, 1]])
    >>> a
    array([[10,  7,  4],
           [ 3,  2,  1]])
    >>> bn.median(a)
    3.5
    >>> bn.median(a, axis=0)
    array([ 6.5,  4.5,  2.5])
    >>> bn.median(a, axis=1)
    array([ 7.,  2.])
    
    """
    func, arr = median_selector(arr, axis)
    return func(arr)

def median_selector(arr, axis):
    """
    Return median function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in
    bn.median() is in checking that `axis` is within range, converting `arr`
    into an array (if it is not already an array), and selecting the function
    to use to calculate the mean.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}
        Axis along which the median is to be computed.
    
    Returns
    -------
    func : function
        The median function that matches the number of dimensions and dtype
        of the input array and the axis along which you wish to find the
        median.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine the median of `arr` along axis=0:

    >>> func, a = bn.func.median_selector(arr, axis=0)
    >>> func
    <function median_1d_float64_axis0>
    
    Use the returned function and array to determine the median:

    >>> func(a)
    2.0

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:    
        a = np.array(arr, copy=False)
    cdef tuple key
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if axis is not None:
        if axis < 0:
            axis += ndim
    else:
        a = PyArray_Ravel(a, NPY_CORDER)
        axis = 0
        ndim = 1
    key = (ndim, dtype, axis)
    try:
        func = median_dict[key]
    except KeyError:
        if (axis < 0) or (axis >= ndim):
            raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = median_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a

'''   
"nanmedian template"

from copy import deepcopy
import bottlechest as bn

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# loops ---------------------------------------------------------------------

loop = {}
loop[1] = """\
    if nINDEX0 == 0:
        return np.FLOAT(NAN)
    j = nAXIS - 1
    flag = 1
    for i in range(nINDEX0):
        if b[i] != b[i]:
            while b[j] != b[j]:
                if j <= 0:
                    break
                j -= 1    
            if i >= j:
                flag = 0
                break
            tmp = b[i]
            b[i] = b[j]
            b[j] = tmp
    n = i + flag
    k = n >> 1
    l = 0
    r = n - 1 
    with nogil:       
        while l < r:
            x = b[k]
            i = l
            j = r
            while 1:
                while b[i] < x: i += 1
                while x < b[j]: j -= 1
                if i <= j:
                    tmp = b[i]
                    b[i] = b[j]
                    b[j] = tmp
                    i += 1
                    j -= 1
                if i > j: break
            if j < k: l = i
            if k < i: r = j
    if n % 2 == 0:        
        amax = MINDTYPE
        allnan = 1
        for i in range(k):
            ai = b[i]
            if ai >= amax:
                amax = ai
                allnan = 0
        if allnan == 0:        
            return np.FLOAT(0.5 * (b[k] + amax))
        else:
            return np.FLOAT(b[k])
    else:
        return np.FLOAT(b[k])
"""        
loop[2] = """\
    if nINDEX1 == 0:
        PyArray_FillWithScalar(y, NAN)
        return y
    for iINDEX0 in range(nINDEX0):
        j = nAXIS - 1
        flag = 1
        for i in range(nINDEX1):
            if b[INDEXREPLACE|i|] != b[INDEXREPLACE|i|]:
                while b[INDEXREPLACE|j|] != b[INDEXREPLACE|j|]:
                    if j <= 0:
                        break
                    j -= 1    
                if i >= j:
                    flag = 0
                    break
                tmp = b[INDEXREPLACE|i|]
                b[INDEXREPLACE|i|] = b[INDEXREPLACE|j|]
                b[INDEXREPLACE|j|] = tmp
        n = i + flag
        k = n >> 1
        l = 0
        r = n - 1
        while l < r:
            x = b[INDEXREPLACE|k|]
            i = l
            j = r
            while 1:
                while b[INDEXREPLACE|i|] < x: i += 1
                while x < b[INDEXREPLACE|j|]: j -= 1
                if i <= j:
                    tmp = b[INDEXREPLACE|i|]
                    b[INDEXREPLACE|i|] = b[INDEXREPLACE|j|]
                    b[INDEXREPLACE|j|] = tmp
                    i += 1
                    j -= 1
                if i > j: break
            if j < k: l = i
            if k < i: r = j
        if n % 2 == 0:        
            amax = MINDTYPE
            allnan = 1
            for i in range(k):
                ai = b[INDEXREPLACE|i|]
                if ai >= amax:
                    amax = ai
                    allnan = 0
            if allnan == 0:        
                y[INDEXPOP] = 0.5 * (b[INDEXREPLACE|k|] + amax)
            else:
                y[INDEXPOP] = CASTb[INDEXREPLACE|k|]         
        else:
            y[INDEXPOP] = CASTb[INDEXREPLACE|k|]         
    return y
"""

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = False
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Median of NDIMd array with dtype=DTYPE along axis=AXIS."
    cdef int allnan = 1, flag = 0
    cdef np.npy_intp i = 0, j = 0, l, r, k, n 
    cdef np.DTYPE_t x, tmp, amax, ai
    cdef np.ndarray[np.DTYPE_t, ndim=NDIM] b = PyArray_Copy(a)
"""

floats['loop'] = {}
floats['loop'][1] = loop[1].replace('FLOAT', 'DTYPE')
floats['loop'][2] = loop[2].replace('CAST', '')

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 
ints['reuse_non_nan_func'] = True

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "nanmedian"
slow['signature'] = "arr"
slow['func'] = "bn.slow.nanmedian(arr, axis=AXIS)"

# Template ------------------------------------------------------------------

nanmedian = {}
nanmedian['name'] = 'nanmedian'
nanmedian['is_reducing_function'] = True
nanmedian['cdef_output'] = True
nanmedian['slow'] = slow
nanmedian['templates'] = {}
nanmedian['templates']['float'] = floats
nanmedian['templates']['int'] = ints
nanmedian['pyx_file'] = 'func/%sbit/median.pyx'

nanmedian['main'] = '''"nanmedian auto-generated from template"
# Select smallest k elements code used for inner loop of median method:
# http://projects.scipy.org/numpy/attachment/ticket/1213/quickselect.pyx
# (C) 2009 Sturla Molden
# SciPy license 
#
# From the original C function (code in public domain) in:
#   Fast median search: an ANSI C implementation
#   Nicolas Devillard - ndevilla AT free DOT fr
#   July 1998
# which, in turn, took the algorithm from
#   Wirth, Niklaus
#   Algorithms + data structures = programs, p. 366
#   Englewood Cliffs: Prentice-Hall, 1976
#
# Adapted and expanded for Bottleneck:
# (C) 2010 Keith Goodman

def nanmedian(arr, axis=None):
    """
    Median of array elements along given axis ignoring NaNs.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}, optional
        Axis along which the median is computed. The default (axis=None) is to
        compute the median of the flattened array.

    Returns
    -------
    y : ndarray
        An array with the same shape as `arr`, except that the specified axis
        has been removed. If `arr` is a 0d array, or if axis is None, a scalar
        is returned. `float64` return values are used for integer inputs. 
    
    See also
    --------
    bottlechest.median: Median along specified axis. 

    Examples
    --------
    >>> a = np.array([[np.nan, 7, 4], [3, 2, 1]])
    >>> a 
    array([[ nan,   7.,   4.],
           [  3.,   2.,   1.]])
    >>> bn.nanmedian(a)
    3.0
    >> bn.nanmedian(a, axis=0)
    array([ 3. ,  4.5,  2.5])
    >> bn.nanmedian(a, axis=1)
    array([ 5.5,  2. ])
    
    """
    func, arr = nanmedian_selector(arr, axis)
    return func(arr)

def nanmedian_selector(arr, axis):
    """
    Return nanmedian function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in
    bn.nanmedian() is in checking that `axis` is within range, converting `arr`
    into an array (if it is not already an array), and selecting the function
    to use to calculate the mean.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}
        Axis along which the median is to be computed.
    
    Returns
    -------
    func : function
        The nanmedian function that matches the number of dimensions and dtype
        of the input array and the axis along which you wish to find the
        median.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine the median of `arr` along axis=0:

    >>> func, a = bn.func.nanmedian_selector(arr, axis=0)
    >>> func
    <function nanmedian_1d_float64_axis0>
    
    Use the returned function and array to determine the median:

    >>> func(a)
    2.0

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:    
        a = np.array(arr, copy=False)
    cdef tuple key
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if axis is not None:
        if axis < 0:
            axis += ndim
    else:
        a = PyArray_Ravel(a, NPY_CORDER)
        axis = 0
        ndim = 1
    key = (ndim, dtype, axis)
    try:
        func = nanmedian_dict[key]
    except KeyError:
        if (axis < 0) or (axis >= ndim):
            raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = nanmedian_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a
'''   




# For support of python 2.5
from __future__ import absolute_import, with_statement

import numpy as np
import bottlechest as bn
from .autotimeit import autotimeit

__all__ = ['bench']


def bench(mode='fast', dtype='float64', axis=1,
          shapes=[(10,10),(100,100),(1000,1000),(10,10),(100,100),(1000,1000)],
          nans=[False, False, False, True, True, True]):
    """
    Bottleneck benchmark.

    Parameters
    ----------
    mode : {'fast', 'faster'}, optional
        Whether to benchmark the high-level functions ('fast') such as
        bn.median or the low-level functions ('faster') such as
        bn.func.median_2d_float64_axis0. By default the high-level functions
        are benchmarked.
    dtype : str, optional
        Data type string such as 'float64', which is the default.
    axis : int, optional
        Axis along which to perform the calculations that are being
        benchmarked. The default is the second axis (axis=1).
    shapes : list, optional
        A list of tuple shapes of input arrays to use in the benchmark.
    nans : list, optional
        A list of the bools (True or False), one for each tuple in the
        `shapes` list, that tells whether the input arrays should be filled
        with one-third NaNs.

    Returns
    -------
    A benchmark report is printed to stdout.

    """
    
    try:
        import scipy as sp
        SCIPY = True
    except ImportError:
        SCIPY = False

    if len(shapes) != len(nans):
        raise ValueError("`shapes` and `nans` must have the same length")

    dtype = str(dtype)
    axis = str(axis)

    tab = '    '

    # Header
    print('Bottleneck performance benchmark')
    print("%sBottleneck  %s" % (tab, bn.__version__))
    print("%sNumpy (np)  %s" % (tab, np.__version__))
    if SCIPY:
        print("%sScipy (sp)  %s" % (tab, sp.__version__))
    else:
        print("%sScipy (sp)  Cannot import, skipping scipy benchmarks" % tab)
    print("%sSpeed is NumPy or SciPy time divided by Bottleneck time" % tab)
    tup = (tab, dtype, axis)
    print("%sNaN means one-third NaNs; %s and axis=%s are used" % tup )
    if mode == 'fast':
        print("%sHigh-level functions used (mode='fast')" % tab)
    elif mode == 'faster':    
        print("%sLow-level functions used (mode='faster')" % tab)
    
    print('')
    header = [" "*14]
    for nan in nans:
        if nan:
            header.append("NaN".center(11))
        else:
            header.append("no NaN".center(11))
    print("".join(header))
    header = ["".join(str(shape).split(" ")).center(11) for shape in shapes]
    header = [" "*14] + header
    print("".join(header))

    suite = benchsuite(mode, shapes, dtype, axis, nans)
    for test in suite:
        name = test["name"].ljust(12)
        fmt = name + "%10.2f" + "%11.2f"*(len(shapes) - 1)
        if test['scipy_required'] and not SCIPY:
            print("%s%s" % (name, "requires SciPy"))
        else:
            speed = timer(test['statements'], test['setups'])
            print(fmt % tuple(speed))

    print('')
    print('Reference functions:')
    for test in suite:
        print("%s%s" % (test["name"].ljust(15), test['ref']))

def timer(statements, setups):
    speed = []
    if len(statements) != 2:
        raise ValueError("Two statements needed.")
    for setup in setups:
        with np.errstate(invalid='ignore'):
            t0 = autotimeit(statements[0], setup)
            t1 = autotimeit(statements[1], setup)
        speed.append(t1 / t0)
    return speed

def getarray(shape, dtype, nans=False):
    arr = np.arange(np.prod(shape), dtype=dtype)
    if nans and issubclass(arr.dtype.type, np.inexact):
        arr[::3] = np.nan
    else:
        rs = np.random.RandomState(shape)
        rs.shuffle(arr)
    return arr.reshape(*shape)
    
def benchsuite(mode, shapes, dtype, axis, nans):

    if mode not in ('fast', 'faster'):
        raise ValueError("`mode` must be 'fast' or 'faster'")

    suite = []
   
    def getsetups(setup, shapes, nans):
        template = """import numpy as np
        import bottlechest as bn
        from bottlechest.benchmark.bench import getarray
        a = getarray(%s, 'DTYPE', %s)
        %s"""
        setups = []
        for shape, nan in zip(shapes, nans):
            setups.append(template % (str(shape), str(nan), setup))
        return setups

    # median
    run = {}
    run['name'] = "median"
    run['ref'] = "np.median"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.median(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "np.median(a, axis=AXIS)"] 
    setup = """
        func, a = bn.func.median_selector(a, axis=AXIS)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # nanmedian
    run = {}
    run['name'] = "nanmedian"
    run['ref'] = "local copy of sp.stats.nanmedian"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nanmedian(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "scipy_nanmedian(a, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.func import scipy_nanmedian
        func, a = bn.func.nanmedian_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # nansum
    run = {}
    run['name'] = "nansum"
    run['ref'] = "np.nansum"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nansum(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "np.nansum(a, axis=AXIS)"] 
    setup = """
        func, a = bn.func.nansum_selector(a, axis=AXIS)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)

    # nanmax
    run = {}
    run['name'] = "nanmax"
    run['ref'] = "np.nanmax"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nanmax(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "np.nanmax(a, axis=AXIS)"] 
    setup = """
        func, a = bn.func.nanmax_selector(a, axis=AXIS)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # nanmean
    run = {}
    run['name'] = "nanmean"
    run['ref'] = "local copy of sp.stats.nanmean"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nanmean(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "scipy_nanmean(a, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.func import scipy_nanmean
        func, a = bn.func.nanmean_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)

    # nanstd
    run = {}
    run['name'] = "nanstd"
    run['ref'] = "local copy of sp.stats.nanstd"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nanstd(a, axis=AXIS)"
    else:
        code = "func(a, 0)"
    run['statements'] = [code, "scipy_nanstd(a, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.func import scipy_nanstd
        func, a = bn.func.nanstd_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # nanargmax
    run = {}
    run['name'] = "nanargmax"
    run['ref'] = "np.nanargmax"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.nanargmax(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "np.nanargmax(a, axis=AXIS)"] 
    setup = """
        func, a = bn.func.nanargmax_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)

    # ss
    run = {}
    run['name'] = "ss"
    run['ref'] = "scipy.stats.ss"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.ss(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "scipy_ss(a, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.func import scipy_ss
        func, a = bn.func.ss_selector(a, axis=AXIS)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # rankdata
    run = {}
    run['name'] = "rankdata"
    run['ref'] = "scipy.stats.rankdata based (axis support added)"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.rankdata(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "bn.slow.rankdata(a, axis=AXIS)"] 
    setup = """
        ignore = bn.slow.rankdata(a, axis=AXIS)
        func, a = bn.func.rankdata_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # partsort
    run = {}
    run['name'] = "partsort"
    run['ref'] = "np.sort, n=max(a.shape[%s]/2,1)" % axis
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.partsort(a, n=n, axis=AXIS)"
    else:
        code = "func(a, n)"
    run['statements'] = [code, "np.sort(a, axis=AXIS)"] 
    setup = """
        if AXIS is None: n = a.size
        else: n = a.shape[AXIS]
        n = max(n / 2, 1)
        func, a = bn.func.partsort_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # argpartsort
    run = {}
    run['name'] = "argpartsort"
    run['ref'] = "np.argsort, n=max(a.shape[%s]/2,1)" % axis
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.argpartsort(a, n=n, axis=AXIS)"
    else:
        code = "func(a, n)"
    run['statements'] = [code, "np.argsort(a, axis=AXIS)"] 
    setup = """
        if AXIS is None: n = a.size
        else: n = a.shape[AXIS]
        n = max(n / 2, 1)
        func, a = bn.func.argpartsort_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)

    # replace
    run = {}
    run['name'] = "replace"
    run['ref'] = "np.putmask based (see bn.slow.replace)"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.replace(a, np.nan, 0)"
    else:
        code = "func(a, np.nan, 0)"
    run['statements'] = [code, "replace(a, np.nan, 0)"] 
    setup = """
        from bottlechest.slow.func import replace
        func = bn.func.replace_selector(a)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)
    
    # anynan
    run = {}
    run['name'] = "anynan"
    run['ref'] = "np.isnan(arr).any(axis)"
    run['scipy_required'] = False
    if mode == 'fast':
        code = "bn.anynan(a, axis=AXIS)"
    else:
        code = "func(a)"
    run['statements'] = [code, "np.isnan(a).any(axis=AXIS)"] 
    setup = """
        func, a = bn.func.anynan_selector(a, axis=AXIS)
    """    
    run['setups'] = getsetups(setup, shapes, nans)
    suite.append(run)

    # move_sum
    run = {}
    run['name'] = "move_sum"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_sum(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_sum(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_sum as scipy_move_sum
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_sum(a, window=w, axis=AXIS)
        func, a = bn.move.move_sum_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_nansum
    run = {}
    run['name'] = "move_nansum"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_nansum(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_nansum(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_nansum as scipy_move_nansum
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_nansum(a, window=w, axis=AXIS)
        func, a = bn.move.move_nansum_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_mean
    run = {}
    run['name'] = "move_mean"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_mean(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_mean(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_mean as scipy_move_mean
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_mean(a, window=w, axis=AXIS)
        func, a = bn.move.move_mean_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_nanmean
    run = {}
    run['name'] = "move_nanmean"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_nanmean(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_nanmean(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_nanmean as scipy_move_nanmean
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_nanmean(a, window=w, axis=AXIS)
        func, a = bn.move.move_nanmean_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_std
    run = {}
    run['name'] = "move_std"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_std(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w, 0)"
    run['statements'] = [code, "scipy_move_std(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_std as scipy_move_std
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_std(a, window=w, axis=AXIS)
        func, a = bn.move.move_std_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_nanstd
    run = {}
    run['name'] = "move_nanstd"
    run['ref'] = "sp.ndimage.convolve1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_nanstd(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w, 0)"
    run['statements'] = [code, "scipy_move_nanstd(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_nanstd as scipy_move_nanstd
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_nanstd(a, window=w, axis=AXIS)
        func, a = bn.move.move_nanstd_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_max
    run = {}
    run['name'] = "move_max"
    run['ref'] = "sp.ndimage.maximum_filter1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_max(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_max(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_max as scipy_move_max
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_max(a, window=w, axis=AXIS)
        func, a = bn.move.move_max_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # move_nanmax
    run = {}
    run['name'] = "move_nanmax"
    run['ref'] = "sp.ndimage.maximum_filter1d based, "
    run['ref'] += "window=a.shape[%s] // 5" % axis
    run['scipy_required'] = True
    if mode == 'fast':
        code = "bn.move_nanmax(a, window=w, axis=AXIS)"
    else:
        code = "func(a, w)"
    run['statements'] = [code, "scipy_move_nanmax(a, window=w, axis=AXIS)"] 
    setup = """
        from bottlechest.slow.move import move_nanmax as scipy_move_nanmax
        w = a.shape[AXIS] // 5
        ignore = bn.slow.move_nanmax(a, window=w, axis=AXIS)
        func, a = bn.move.move_nanmax_selector(a, axis=AXIS)
    """
    run['setups'] = getsetups(setup, shapes, nans)
    if axis != 'None':
        suite.append(run)
    
    # Strip leading spaces from setup code
    for i, run in enumerate(suite):
        for j in range(len(run['setups'])):
            t = run['setups'][j]
            t = '\n'.join([z.strip() for z in t.split('\n')])
            suite[i]['setups'][j] = t
            
    # Set dtype and axis in setups
    for i, run in enumerate(suite):
        for j in range(len(run['setups'])):
            t = run['setups'][j]
            t = t.replace('DTYPE', dtype)
            t = t.replace('AXIS', axis)
            suite[i]['setups'][j] = t

    # Set dtype and axis in statements
    for i, run in enumerate(suite):
        for j in range(2):
            t = run['statements'][j]
            t = t.replace('DTYPE', dtype)
            t = t.replace('AXIS', axis)
            suite[i]['statements'][j] = t

    return suite 



import unittest
import sys
import bottlechest as bn
import numpy as np
import scipy.sparse as sp


class TestStats(unittest.TestCase):
    def test_simple_int(self):
        data = np.array([[0, 1, 1, 2, 0],
                         [1, 1, 1, 0, 0],
                         [0, 0, 3, 0, 0]], dtype=int)
        y = bn.stats(data)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 0, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 0])
        np.testing.assert_almost_equal(y[:, 2], [1/3, 2/3, 5/3, 2/3, 0])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 5], [3, 3, 3, 3, 3])

        y = bn.stats(data, None, True)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 0, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 0])
        np.testing.assert_almost_equal(y[:, 2], [1/3, 2/3, 5/3, 2/3, 0])
        np.testing.assert_almost_equal(y[:, 3], [1/3, 1/3, 4/3, 4/3, 0])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 5], [3, 3, 3, 3, 3])

    def test_simple_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 0],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        y = bn.stats(data)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 2, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [1/3, 2/3, 5/3, 2, 0.5])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 2, 1])
        np.testing.assert_equal(y[:, 5], [3, 3, 3, 1, 2])

        y = bn.stats(data, None, True)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 2, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [1/3, 2/3, 5/3, 2, 0.5])
        np.testing.assert_almost_equal(y[:, 3], [1/3, 1/3, 4/3, 0, 1/2])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 2, 1])
        np.testing.assert_equal(y[:, 5], [3, 3, 3, 1, 2])

    def test_simple_float_allnan(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, nan, 0],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        y = bn.stats(data)
        np.testing.assert_equal(y[3, :],
            [float("inf"), float("-inf"), 0, 0, 3, 0])

        y = bn.stats(data, None, True)
        np.testing.assert_equal(y[3, :],
            [float("inf"), float("-inf"), 0, 0, 3, 0])

    def test_weighted_int(self):
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, 0, 1],
                         [0, 0, 3, 0, 0]], dtype=int)
        y = bn.stats(data, np.array([1, 2, 3], dtype=float))
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 0, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [2/6, 0.5, 2, 2/6, 0.5])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 5], [6, 6, 6, 6, 6])

        y = bn.stats(data, np.array([1, 2, 3], dtype=float), True)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 0, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [2/6, 0.5, 2, 2/6, 0.5])
        np.testing.assert_almost_equal(
            y[:, 3], [(1/9 + 2 * 4/9 + 3 * 1/9) * 6 / 22,
                      0.25 * 6 * 6 / 22,
                      6 * 6 / 22,
                      (25/9 + 2 * 1/9 + 3 * 1/9) * 6 / 22,
                      0.25 * 6 * 6 / 22])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 5], [6, 6, 6, 6, 6])

    def test_weighted_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 0],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        y = bn.stats(data, np.array([1, 2, 3], dtype=float))
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 2, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [2/6, 0.5, 2, 2, 2/3])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 5, 3])
        np.testing.assert_equal(y[:, 5], [6, 6, 6, 1, 3])

        y = bn.stats(data, np.array([1, 2, 3], dtype=float), True)
        np.testing.assert_equal(y[:, 0], [0, 0, 1, 2, 0])
        np.testing.assert_equal(y[:, 1], [1, 1, 3, 2, 1])
        np.testing.assert_almost_equal(y[:, 2], [2/6, 0.5, 2, 2, 2/3])
        np.testing.assert_almost_equal(
            y[:, 3], [(1/9 + 2 * 4/9 + 3 * 1/9) * 6 / 22,
                      0.25 * 6 * 6 / 22,
                      6 * 6 / 22,
                      0,
                      (4/9 + 2 * 1/9) * 3 / 4])
        np.testing.assert_equal(y[:, 4], [0, 0, 0, 5, 3])
        np.testing.assert_equal(y[:, 5], [6, 6, 6, 1, 3])

    # 0 1 2 3
    #--------
    # 1 1 2
    # 2
    #   1 3
    def test_sparse_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))

        y = bn.stats(a)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [1.5, 1, 2.5, 0])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [1, 1, 1, 3])
        np.testing.assert_equal(y[:, 5], [2, 2, 2, 0])

        y = bn.stats(a, None, True)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [1.5, 1, 2.5, 0])
        np.testing.assert_equal(y[:, 3], [0.5, 0, 0.5, 0])
        np.testing.assert_equal(y[:, 4], [1, 1, 1, 3])
        np.testing.assert_equal(y[:, 5], [2, 2, 2, 0])

    def test_sparse_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        y = bn.stats(a)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [1.5, 1, 2.5, 0])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [1, 1, 1, 3])
        np.testing.assert_equal(y[:, 5], [2, 2, 2, 0])

        y = bn.stats(a, None, True)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [1.5, 1, 2.5, 0])
        np.testing.assert_almost_equal(y[:, 3], [0.5, 0, 0.5, 0])
        np.testing.assert_equal(y[:, 4], [1, 1, 1, 3])
        np.testing.assert_equal(y[:, 5], [2, 2, 2, 0])

    # 0 1 2 3  w
    #--------
    # 1 1 2    1
    # 2        2
    #   1 3    3
    def test_sparse_weight_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        y = bn.stats(a, [1, 2, 3])
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [5/3, 1, 2.75, 0])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [3, 2, 2, 6])
        np.testing.assert_equal(y[:, 5], [3, 4, 4, 0])

        y = bn.stats(a, [1, 2, 3], True)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [5/3, 1, 2.75, 0])
        np.testing.assert_almost_equal(y[:, 3], [(4/9 + 2 * 1/9) * 3/4, 0,
                                                 (9/16 + 3 * 1/16) * 4/6, 0])
        np.testing.assert_equal(y[:, 4], [3, 2, 2, 6])
        np.testing.assert_equal(y[:, 5], [3, 4, 4, 0])

    def test_sparse_weight_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        y = bn.stats(a, [1, 2, 3])
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [5/3, 1, 2.75, 0])
        np.testing.assert_equal(y[:, 3], [0, 0, 0, 0])
        np.testing.assert_equal(y[:, 4], [3, 2, 2, 6])
        np.testing.assert_equal(y[:, 5], [3, 4, 4, 0])

        y = bn.stats(a, [1, 2, 3], True)
        np.testing.assert_equal(y[:, 0], [1, 1, 2, float("inf")])
        np.testing.assert_equal(y[:, 1], [2, 1, 3, float("-inf")])
        np.testing.assert_almost_equal(y[:, 2], [5/3, 1, 2.75, 0])
        np.testing.assert_almost_equal(y[:, 3], [(4/9 + 2 * 1/9) * 3/4, 0,
                                                 (9/16 + 3 * 1/16) * 4/6, 0])
        np.testing.assert_equal(y[:, 4], [3, 2, 2, 6])
        np.testing.assert_equal(y[:, 5], [3, 4, 4, 0])


    def test_int_1d(self):
        data = np.array([0, 1, 1, 2, 0])
        a_min, a_max, mean, var, nans, non_nans = bn.stats(data)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 4/5)
        self.assertEqual(var, 0)
        self.assertEqual(nans, 0)
        self.assertEqual(non_nans, 5)

        a_min, a_max, mean, var, nans, non_nans = bn.stats(data, None, True)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 4/5)
        self.assertAlmostEqual(var, (16 + 1 + 1 + 36 + 16) / 25 / 4)
        self.assertEqual(nans, 0)
        self.assertEqual(non_nans, 5)

    def test_float_1d(self):
        data = np.array([0, 1, float("nan"), 2, 0], dtype="float")
        a_min, a_max, mean, var, nans, non_nans = bn.stats(data)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 3/4)
        self.assertEqual(var, 0)
        self.assertEqual(nans, 1)
        self.assertEqual(non_nans, 4)

        a_min, a_max, mean, var, nans, non_nans = bn.stats(data, None, True)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 3/4)
        self.assertEqual(var, (9/16 + 1/16 + 25/16 + 9/16) / 3)
        self.assertEqual(nans, 1)
        self.assertEqual(non_nans, 4)

    def test_int_weight_1d(self):
        data = np.array([0, 1, 1, 2, 0])
        a_min, a_max, mean, var, nans, non_nans = \
            bn.stats(data, [1, 2, 3, 4, 5])
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 13/15)
        self.assertEqual(var, 0)
        self.assertEqual(nans, 0)
        self.assertEqual(non_nans, 15)

        a_min, a_max, mean, var, nans, non_nans = \
            bn.stats(data, [1, 2, 3, 4, 5], True)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 13/15)
        self.assertEqual(var, (169 + 2*4 + 3*4 + 4*17**2 + 5*169) / 225 * 15/170)
        self.assertEqual(nans, 0)
        self.assertEqual(non_nans, 15)

    def test_float_weight_1d(self):
        data = np.array([0, 1, float("nan"), 2, 0], dtype="float")
        a_min, a_max, mean, var, nans, non_nans = \
            bn.stats(data, [1, 2, 3, 4, 5])
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 10/12)
        self.assertEqual(var, 0)
        self.assertEqual(nans, 3)
        self.assertEqual(non_nans, 12)

        a_min, a_max, mean, var, nans, non_nans = \
            bn.stats(data, [1, 2, 3, 4, 5], True)
        self.assertEqual(a_min, 0)
        self.assertEqual(a_max, 2)
        self.assertAlmostEqual(mean, 10/12)
        self.assertAlmostEqual(var, (100 + 2*4 + 4*196 + 5*100) / 144 * 12/98)
        self.assertEqual(nans, 3)
        self.assertEqual(non_nans, 12)

if __name__ == "__main__":
    unittest.main()
